#!/usr/bin/python

"""
This module provides a base class for the attribute-carrying objects
and other common utility functions.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'

from warnings import warn
import sys


if sys.version_info[0] >= 3:
    # pylint: disable=redefined-builtin
    # pylint: disable=invalid-name
    unicode = str
    basestring = str


class AttributeCarrier(object):
    """
    Base class that provides utility methods.

    Each derived class has an allowed_args attribute, listing all instance
    attributes that can be set at initialization time by providing them
    as keyword arguments.

    Also, utility methods are provided to check the validity of the attributes
    and emit XML trees for the represented objects.
    """
    # pylint: disable=no-member

    allowed_args = ()  # Always redefined by derived classes.

    def __init__(self, **kwargs):
        self.process(kwargs, self.allowed_args)

    def process(self, kwargs, allowed_args, final=True):
        """
        Set allowed attributes on the object instance. Return dictionary without
        the processed attributes.
        """
        for arg in allowed_args:
            if arg in kwargs:
                setattr(self, arg, kwargs.pop(arg))
        if final and len(kwargs) > 0:
            raise TypeError('Invalid keyword arguments: %s' % kwargs.keys())
        return kwargs

    def __tag__(self, *args, **kwargs):
        """
        Return a XML tag for the object, after performing all the validity
        checks.
        """
        self.perform_checks()
        return self.emit_tag(*args, **kwargs)

    def max_length(self, attribute_name, length, obj=None):
        "Check that an attribute fits into the field length."
        if obj is None:
            obj = self
        value = unicode(getattr(obj, attribute_name))
        if len(value) < 1:
            raise Exception('Attribute %r cannot be empty' % attribute_name)
        if len(value) > length:
            warn('Attribute %r too long; truncating' % attribute_name,
                 stacklevel=4)
            setattr(obj, attribute_name, value[:length])
        else:
            setattr(obj, attribute_name, value)

    def length(self, attribute_name, length):
        "Check that an attribute has exactly the specified length."
        value = unicode(getattr(self, attribute_name))
        if len(value) != length:
            raise Exception(
                'Attribute %r must be %s characters long '
                '(supplied value: %r' % (attribute_name, length, value))
        setattr(self, attribute_name, value)

def booltext(param):
    "Returns a string suitable to represent a boolean value in a XML file."
    if param:
        return 'true'
    else:
        return 'false'



from sepacbi import Payment, Invoice, DebitNote, Text
from .definitions import *
import re
import sys
from datetime import datetime, date

from lxml import etree

PYTHON3 = False
if sys.version_info[0] >= 3:
    PYTHON3 = True


TIMESTAMP_RE = re.compile(r'<CreDtTm>[-0-9:.T]+</CreDtTm>')
CANONICAL_TIMESTAMP = '<CreDtTm>' + datetime.now().isoformat() + '</CreDtTm>'


def canonicalize_xml(xmlstring):
    """
    Wipe out the info that cannot be constant in the XML string (such as
    autogenerated timestamps), parse it and and reoutput it.
    """
    if PYTHON3:
        if hasattr(xmlstring, 'decode'):
            xmlstring = xmlstring.decode(encoding='utf-8')
    string = TIMESTAMP_RE.sub(CANONICAL_TIMESTAMP, xmlstring)
    tree = etree.fromstring(string)
    return etree.tostring(tree, pretty_print=True)


def compare_xml(tree, filename, save=False):
    """
    Compare a generated XML tree with the content of a specific file.
    """
    from os.path import dirname, join
    full_path = join(dirname(__file__), filename)
    tree = canonicalize_xml(tree)
    if save:
        open(full_path, 'w').write(tree)
        raise Exception('saved!')
    file_content = canonicalize_xml(open(full_path, 'r').read())
    assert tree == file_content


def test_payment_basic():
    payment = Payment(debtor=biz_with_cuc, account=acct_37, req_id='StaticId',
                      execution_date=date(2014, 5, 15))
    payment.add_transaction(amount=198.25, account=acct_86, creditor=beta,
                            rmtinfo='Causale 1')
    compare_xml(payment.xml_text(), 'payment_basic.xml')


def test_payment_multitrans():
    payment = Payment(debtor=biz_with_cuc, account=acct_37, req_id='StaticId',
                      execution_date=date(2014, 5, 15))
    payment.add_transaction(amount=198.25, account=acct_86, creditor=beta,
                            rmtinfo='Causale 1')
    payment.add_transaction(amount=350.00, account=acct_37,
                            creditor=biz_with_cuc,
                            rmtinfo='Altra causale')
    payment.add_transaction(amount=9532.21, account=foreign_acct,
                            bic='ABCDESNN', creditor=alpha,
                            docs=[Invoice(18512, 4500),
                                  DebitNote(1048, 5032.21,
                                            date(1995, 4, 21))])
    payment.add_transaction(
        amount=1242.80, creditor=pvt, account=acct_86,
        category='SALA', rmtinfo='Salary')

    compare_xml(payment.xml_text(), 'payment_multitrans.xml')


def test_payment_misc_features():
    payment = Payment(
        debtor=biz, account=acct_37, req_id='StaticId',
        execution_date=date(2014, 5, 15),
        ultimate_debtor=beta, charges_account=acct_86, envelope=True,
        initiator=biz_with_cuc, batch=True, high_priority=True)
    payment.add_transaction(amount=198.25, account=acct_86, creditor=beta,
                            rmtinfo='Causale 1')
    compare_xml(payment.xml_text(), 'payment_misc_1.xml')

    payment.high_priority = False
    payment.add_transaction(amount=350.00, account=acct_37,
                            creditor=biz_with_cuc,
                            rmtinfo='Altra causale')
    payment.add_transaction(amount=9532.21, account=foreign_acct,
                            bic='ABCDESNN', creditor=alpha,
                            docs=[Invoice(18512, 4500),
                                  DebitNote(1048, 5032.21,
                                            date(1995, 4, 21))])
    payment.add_transaction(
        amount=1242.80, creditor=pvt, account=acct_86,
        category='SALA', docs=[Text('Salary payment')])

    compare_xml(payment.xml_text(), 'payment_misc_2.xml')



#!/usr/bin/python

"""
This module emits the root tag for the credit transfer.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'

import sys

from lxml import etree
from .util import AttributeCarrier, booltext
from .entity import IdHolder
from .account import Account
from .bank import Bank
from .transaction import Transaction
from .cbibon_dom import PCRecord, EFRecord
from datetime import date, datetime

if sys.version_info[0] >= 3:
    # pylint: disable=redefined-builtin
    # pylint: disable=invalid-name
    basestring = str


class MissingABIError(Exception):
    """
    Raised when the debtor's ABI is not specified and cannot be inferred from
    the account.
    """


class NoTransactionsError(Exception):
    """
    Raised when trying to generate a credit transfer request containing no
    transactions.
    """
    pass


class InvalidEndToEndIDError(Exception):
    """
    Raised when a transaction has a non-unique end-to-end ID.
    """


class Payment(AttributeCarrier):
    # pylint: disable=no-member
    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=access-member-before-definition
    """
    Payment represents the root tag for the credit transfer and holds
    all the associated information in its attributes, including the list
    of transactions to be performed.
    """
    allowed_args = (
        'req_id', 'batch', 'high_priority', 'execution_date',
        'debtor', 'account', 'abi', 'ultimate_debtor', 'charges_account',
        'envelope', 'initiator')

    ID_PREFIX = 'DistintaXml-'

    def __init__(self, **kwargs):
        self.envelope = False
        self.transactions = []
        self.eeid_set = set()
        super(Payment, self).__init__(**kwargs)

    def add_eeid(self, txid):
        "Add a transaction's end-to-end ID to check for uniqueness."
        if txid in self.eeid_set:
            raise InvalidEndToEndIDError('Duplicate end-to-end ID: %r' % txid)
        self.eeid_set.add(txid)

    def add_transaction(self, **kwargs):
        "Adds a transaction to the internal list. Does not return anything."
        kwargs['payment_seq'] = len(self.transactions)+1
        if not hasattr(self, 'req_id'):
            self.gen_id()
        kwargs['payment_id'] = self.req_id
        kwargs['register_eeid_function'] = self.add_eeid
        kwargs['payment'] = self
        txr = Transaction(**kwargs)
        txr.perform_checks()
        self.transactions.append(txr)

    def gen_id(self):
        """Generate a unique ID for the payment"""
        self.req_id = '%s%s' % (self.ID_PREFIX, datetime.now().strftime(
            '%Y%m%d-%H%M%S'))

    def perform_checks(self):
        "Checks the validity of all supplied attributes."
        if not hasattr(self, 'req_id'):
            self.gen_id()
        self.max_length('req_id', 35)

        assert isinstance(self.debtor, IdHolder)

        if isinstance(self.account, basestring):
            self.account = Account(iban=self.account)
        assert isinstance(self.account, Account)

        if hasattr(self, 'abi'):
            abi = self.abi
        elif self.account.iban[:2] == 'IT':
            abi = self.account.iban[5:10]
        else:
            raise MissingABIError('The payment needs an \'abi\' attribute')
        self.bank = Bank(abi=abi)

        # for CBI text files
        self.cab = self.account.iban[10:15]
        self.cc = self.account.iban[15:27]

        if hasattr(self, 'ultimate_debtor'):
            assert isinstance(self.ultimate_debtor, IdHolder)

        if hasattr(self, 'charges_account'):
            if isinstance(self.charges_account, basestring):
                self.charges_account = Account(iban=self.charges_account)
            assert isinstance(self.charges_account, Account)

        # Todo: if there is an initiator, check that it has a CUC
        # Todo: if there is no initiator, check that the debtor has a CUC

    def amount_sum(self):
        return sum([tx.amount for tx in self.transactions])

    def get_initiator(self):
        """
        Returns the entity designated as initiator for the transfer.
        """
        if hasattr(self, 'initiator'):
            return self.initiator
        else:
            return self.debtor

    def get_xml_root(self):
        """
        Builds the XML structure for the payment instruction, as dictated by the
        relevant standard.

        Returns a pair consisting of the root tag and the tag to which the
        Payment information should be appended. This allows the function to
        pick a different outer structure according to the value of the
        `envelope` attribute.
        """
        xsi = 'http://www.w3.org/2001/XMLSchema-instance'
        if self.envelope:
            tag = 'CBIBdyPaymentRequest'
        else:
            tag = 'CBIPaymentRequest'
        schema = tag + '.00.04.00'
        xmlns = 'urn:CBI:xsd:' + schema
        schema_location = xmlns + ' ' + schema + '.xsd'
        root = etree.Element(
            '{%s}%s' % (xmlns, tag),
            attrib={'{%s}schemaLocation' % xsi: schema_location},
            nsmap={'xsi': xsi, None: xmlns})
        outer = root
        if self.envelope:
            root = etree.SubElement(root, 'CBIEnvelPaymentRequest')
            root = etree.SubElement(root, 'CBIPaymentRequest')
        return outer, root

    def emit_tag(self):
        "Returns the whole XML structure for the payment."
        # Outer XML structure
        outer, root = self.get_xml_root()
        xmlns = 'urn:CBI:xsd:CBIPaymentRequest.00.04.00'

        # Header
        header = etree.SubElement(root, 'GrpHdr', nsmap={None: xmlns})
        etree.SubElement(header, 'MsgId').text = self.req_id
        etree.SubElement(header, 'CreDtTm').text = datetime.now().isoformat()
        etree.SubElement(header, 'NbOfTxs').text = str(len(self.transactions))
        etree.SubElement(header, 'CtrlSum').text = str(self.amount_sum())
        initiator = self.get_initiator()
        header.append(initiator.__tag__(as_initiator=True))

        # Payment info
        info = etree.SubElement(root, 'PmtInf', nsmap={None: xmlns})

        # TRF: no status requested
        etree.SubElement(info, 'PmtInfId').text = self.req_id
        etree.SubElement(info, 'PmtMtd').text = 'TRF'

        # Batch booking
        if hasattr(self, 'batch'):
            etree.SubElement(info, 'BtchBookg').text = booltext(self.batch)

        # Priority
        if hasattr(self, 'high_priority'):
            tp_info = etree.SubElement(info, 'PmtTpInf')
            if hasattr(self, 'high_priority'):
                priority_text = 'NORM'
                if self.high_priority:
                    priority_text = 'HIGH'
                etree.SubElement(tp_info, 'InstrPrty').text = priority_text
            svclvl = etree.SubElement(tp_info, 'SvcLvl')
            etree.SubElement(svclvl, 'Cd').text = 'SEPA'

        # Execution date: either today or specified date
        execution_date = date.today()
        if hasattr(self, 'execution_date'):
            execution_date = self.execution_date
        etree.SubElement(info, 'ReqdExctnDt').text = execution_date.isoformat()

        # Debtor information
        info.append(self.debtor.__tag__('Dbtr'))

        # Debtor account
        info.append(self.account.__tag__('DbtrAcct'))

        agent = etree.SubElement(info, 'DbtrAgt')
        agent.append(self.bank.__tag__(output_abi=True))

        # Ultimate debtor
        if hasattr(self, 'ultimate_debtor'):
            info.append(self.ultimate_debtor.__tag__('UltmtDbtr'))
        etree.SubElement(info, 'ChrgBr').text = 'SLEV'

        # Charges account
        if hasattr(self, 'charges_account'):
            info.append(self.charges_account.__tag__('ChrgsAcct'))

        # Transactions
        if len(self.transactions) == 0:
            raise NoTransactionsError
        for txr in self.transactions:
            info.append(txr.__tag__())

        return outer

    def xml(self):
        """
        Return the lxml tree.
        """
        return self.__tag__()

    def xml_text(self, **kwargs):
        """
        Return the XML structure as a string.
        """
        return etree.tostring(self.xml(), **kwargs)

    def cbi_text(self):
        self.perform_checks()

        if self.account.is_foreign():
            raise Exception('Cannot use foreign accounts with CBI text files')

        today = date.today()

        header = PCRecord()
        footer = EFRecord()

        sia_code = self.get_initiator().sia_code
        header.sender = sia_code
        footer.sender = sia_code
        header.recipient = self.bank.abi
        footer.recipient = self.bank.abi
        header.creation = today
        footer.creation = today
        header.name = 'Distinta'
        footer.name = 'Distinta'
        if hasattr(self, 'high_priority'):
            if self.high_priority:
                header.prio = 'U'
                footer.prio = 'U'

        footer.orders = len(self.transactions)
        footer.negative_amounts = 0
        footer.positive_amounts = self.amount_sum()

        records = [header.format()]
        for transaction, i in zip(self.transactions,
                                  range(len(self.transactions))):
            records += transaction.cbi_records(i+1)
        footer.records = len(records)+1
        records.append(footer.format())
        records.append(u'')
        return '\n'.join(records)






#!/usr/bin/python

"""
Bank accounts.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'

from .util import AttributeCarrier
from lxml import etree


class MissingCUCError(Exception):
    """
    Raised when an Entity without a `cuc` attribute is used as an initiator.
    """
    pass


class AddressFormatError(Exception):
    """
    Raised when an Address is created with unsuitable arguments.
    """


class Address(AttributeCarrier):
    """
    The postal address of an entity.
    """
    def __init__(self, *args):
        """
        The argument processing is NOT based on keyword parameters, so we
        cannot inherit from AttributeCarrier.
        """
        # pylint: disable=super-init-not-called
        self.lines = args

    def perform_checks(self):
        "Check the length of each line."
        line_count = len(self.lines)
        if line_count < 1 or line_count > 2:
            raise AddressFormatError('Must have either 1 or 2 address lines')
        for line in self.lines:
            if len(line) < 1 or len(line) > 70:
                raise AddressFormatError('Line must have length '
                                         'between 1 and 70 characters')

    def emit_tag(self):
        "Emit the postal address with each of its lines."
        root = etree.Element('PstlAdr')
        for line in self.lines:
            etree.SubElement(root, 'AdrLine').text = line
        return root


def emit_id_tag(id_code, id_type=None):
    'Emit a single ID subtree, with root <Othr>.'
    othr = etree.Element('Othr')
    id_tag = etree.SubElement(othr, 'Id')
    id_tag.text = id_code
    if id_type is not None:
        issuer = etree.SubElement(othr, 'Issr')
        issuer.text = id_type
    return othr


class IdHolder(AttributeCarrier):
    # pylint: disable=no-member
    """
    Describes an Initiator.
    Must have a CUC; may have an Italian tax code.
    """
    allowed_args = ('name', 'cf', 'code', 'private', 'cuc', 'address',
                    'country', 'sia_code')

    def __init__(self, **kwargs):
        self.private = False
        super(IdHolder, self).__init__(**kwargs)

    def perform_checks(self):
        # pylint: disable=access-member-before-definition
        # pylint: disable=attribute-defined-outside-init
        "Check argument lengths."
        if hasattr(self, 'name'):
            self.max_length('name', 70)
        if hasattr(self, 'cf'):
            assert not hasattr(self, 'code')
            self.max_length('cf', 16)
        if hasattr(self, 'cuc'):
            self.max_length('cuc', 35)
        if hasattr(self, 'address'):
            if isinstance(self.address, (list, tuple)):
                self.address = Address(*self.address)
            assert isinstance(self.address, Address)
        if hasattr(self, 'country'):
            self.length('country', 2)

    def emit_tag(self, tag=None, as_initiator=False):
        """
        Emit a subtree for an entity, using the supplied tag for the root
        element. If the identity is the Initiator's, emit the CUC as well.
        """
        if as_initiator:
            tag = 'InitgPty'
        root = etree.Element(tag)

        # Name
        if hasattr(self, 'name'):
            name = etree.SubElement(root, 'Nm')
            name.text = self.name

        # Address
        if hasattr(self, 'address') and not as_initiator:
            root.append(self.address.__tag__())

        # ID
        idtag = etree.SubElement(root, 'Id')
        if self.private:
            id_container = 'PrvtId'
        else:
            id_container = 'OrgId'
        orgid = etree.SubElement(idtag, id_container)

        # CUC
        if as_initiator:
            if not hasattr(self, 'cuc'):
                raise MissingCUCError
            orgid.append(emit_id_tag(self.cuc, 'CBI'))

        # Tax code
        if hasattr(self, 'cf'):
            orgid.append(emit_id_tag(self.cf, 'ADE'))

        if hasattr(self, 'code'):
            orgid.append(emit_id_tag(self.code, None))

        if not as_initiator and hasattr(self, 'country'):
            etree.SubElement(root, 'CtryOfRes').text = self.country

        return root



#!/usr/bin/python

"""
This module deals with IBAN codes and formal validity checks.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'

import re
from .iban_structures import IBAN_STRUCTURES

# Identifies a single element of the IBAN structure
STR_ITEM_RE = re.compile(r'^(\d+)(!?)([nac])')


class InvalidIBANError(Exception):
    "Raised when an IBAN does not pass the formal checks."
    pass


def structure_to_re(structure):
    """
    Convert a SWIFT IBAN structure description to a regex that matches it.
    """
    # First part of the regex
    regex = ['^', structure[:2]]
    # Skip the first two characters
    idx = 2
    while idx < len(structure):
        # Get the next segment
        pattern = STR_ITEM_RE.search(structure[idx:])
        idx += len(pattern.group(0))

        # The part length
        length = int(pattern.group(1))

        # Whether it is fixed or a maximum length
        fixed = pattern.group(2) != ''

        # 'n' is a digit; 'a' an uppercase character;
        # 'c' an alphanumeric character
        item_type = pattern.group(3)

        # compose the next regex segment
        if item_type == 'n':
            sub_re = r'\d'
        elif item_type == 'a':
            sub_re = r'[A-Z]'
        elif item_type == 'c':
            sub_re = r'[\dA-Za-z]'
        else:
            raise Exception('Invalid structure: %s' % structure)

        if fixed:
            sub_re += r'{%s}' % length
        else:
            sub_re += r'{,%s}' % length
        regex.append(sub_re)
    regex.append('$')
    regex = ''.join(regex)
    # Return the finished and compiled regex
    return re.compile(regex)


# Fill a dictionary with the compiled regular expressions
COUNTRY_RE = dict([(x[:2], structure_to_re(x))
                   for x in IBAN_STRUCTURES])


def validate_check_digits(iban):
    "Validate the check digits of an IBAN."
    iban = iban[4:] + iban[:4]

    converted = int(''.join(str(int(ch, 36)) for ch in iban))

    if int(converted) % 97 != 1:
        raise InvalidIBANError('Invalid check digits')


def validate(iban):
    "Validate the structure and the check digits of an IBAN."

    # Do we know the country?
    country = iban[:2]
    if country not in COUNTRY_RE:
        raise InvalidIBANError('Invalid country code')

    # Is the formal structure valid?
    if not COUNTRY_RE[country].match(iban):
        raise InvalidIBANError('Invalid IBAN structure for country %s'
                               % country)

    # Are the check digits valid?
    validate_check_digits(iban)



from .records import BaseRecord, AlphaNumericField, NumericField, \
    CBIDateField, DecimalField

from .cbibon_dom import CBIRecord


class PERecord(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default='PE')
        f.nu(4, 'sia', 5)
        f.nu(9, 'abi', 5)
        f.dt(14, 'creation', 6)
        f.an(20, 'name', 20)
        f.an(40, 'available', 6)
        f.an(46, 'filler2', 61)
        f.an(105, 'flow_qualifier', 7)
        f.an(112, 'filler3', 9)


class EFRecord(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default='EF')
        f.an(4, 'sender', 5)
        f.nu(9, 'recipient', 5)
        f.dt(14, 'creation', 6)
        f.an(20, 'name', 20)
        f.an(40, 'available', 6)
        f.nu(46, 'orders', 7)
        f.an(53, 'filler2', 12)
        f.cur(65, 'amounts', 18)
        f.nu(83, 'records', 7)
        f.an(90, 'filler3', 25)
        f.an(115, 'not_available', 6)


class PaymentIdDates(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default='H1')
        f.an(4, 'prog_number', 7)
        f.an(11, 'msg_identifier', 14, default='PAYORDD  93AUN')
        f.an(25, 'msg_name', 3, default='450')
        f.an(28, 'reference', 34)
        f.an(62, 'filler2', 1)
        f.an(63, 'msg_type', 2, default='9 ')
        f.nu(65, 'creation_date_qualifier')
        f.dt(68, 'creation_date', 8)
        f.nu(76, 'creation_date_format', 3, default=102)
        f.nu(79, 'execution_date_qualifier', 3)
        f.dt(82, 'execution_date', 8)
        f.nu(90, 'execution_date_format', 3, default=102)
        f.nu(93, 'availability_date_qualifier', 3)
        f.dt(96, 'availability_date', 8)
        f.nu(104, 'availability_date_format', 3, default=102)
        f.an(107, 'transaction_type', 3, default='IN ')
        f.an(110, 'payment_type', 3)
        f.an(113, 'filler3', 3)
        f.an(116, 'not_available', 5)


class PaymentMode(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default='P0')
        f.an(4, 'prog_number', 7)
        f.an(11, 'payment_form', 3)
        f.an(14, 'charges_split', 3)
        f.an(17, 'charges_account_cab', 5)
        f.an(22, 'charges_account_s1', 3, default='25 ')
        f.an(25, 'charges_account_s2', 3, default='119')
        f.an(28, 'charges_account_code', 17)
        f.an(45, 'charges_account_cin', 1)
        f.an(46, 'filler2', 75)


class PaymentAmount(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default='P1')
        f.an(4, 'prog_number', 7)
        f.an(11, 'currency_relationship', 3, default='9  ')
        f.cur(14, 'amount', 18)
        f.an(32, 'currency', 3, default='EUR')
        f.an(35, 'charge_currency_qualifier', 3)
        f.an(38, 'charge_currency', 3)
        f.an(41, 'charge_exchange_rate', 12)
        f.an(53, 'credit_currency_qualifier', 3)
        f.an(56, 'credit_currency', 3)
        f.an(59, 'credit_exchange_rate', 12)
        f.an(71, 'filler2', 11)
        f.an(82, 'instructions_code', 3)
        f.an(85, 'instructions_contract', 35)
        f.an(120, 'filler3', 1)


class PaymentBanks(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default='PH')
        f.an(4, 'prog_number', 7)
        f.an(11, 'payer_iban', 34)
        f.an(45, 'payer_bic', 11)
        f.an(56, 'filler2', 5)
        f.an(61, 'payee_iban', 34)
        f.an(95, 'payee_bic', 11)
        f.an(106, 'filler3', 15)


class Payee(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default='P5')
        f.an(4, 'prog_number', 7)
        f.an(11, 'fixed', 3, default='BE ')
        f.an(14, 'name', 35)
        f.an(49, 'address', 35)
        f.an(84, 'town', 32)
        f.an(116, 'payee_sia', 5)


class PaymentAmount(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default='P1')
        f.an(4, 'prog_number', 7)











import pytest

from sepacbi import IdHolder, Payment

from sepacbi.payment import NoTransactionsError, MissingABIError, \
    InvalidEndToEndIDError
from sepacbi.entity import MissingCUCError, AddressFormatError
from sepacbi.transaction import MissingBICError
from sepacbi.rmtinfo import Invoice
from sepacbi.iban import InvalidIBANError

from .definitions import *


def test_empty_payment():
    "Payments without any transaction must raise an exception."
    with pytest.raises(NoTransactionsError):
        payment = Payment(
            debtor=biz_with_cuc,
            account=acct_86,
            abi='01101',
            envelope=False
        )
        payment.xml()


def test_missing_attrs_payment():
    "Payments must always have a `debtor` and an `account` attribute."
    payment = Payment()
    with pytest.raises(AttributeError):
        payment.xml()

    payment.debtor = biz_with_cuc
    with pytest.raises(AttributeError):
        payment.xml()

    payment.account = acct_37

    # No exception must be raised
    payment.add_transaction(amount=1, account=acct_86,
                            rmtinfo='Test', creditor=biz)
    payment.xml()


def gen_dictionaries_with_one_missing_item(full_dictionary):
    """
    Takes an input dictionary and yields all possible dictionaries
    with the same data, except for one single key/value pair.
    """
    for item in full_dictionary:
        incomplete = dict([(key, full_dictionary[key])
                           for key in full_dictionary if key != item])
        yield incomplete


def simple_payment():
    return Payment(debtor=biz_with_cuc, account=acct_37)


def test_missing_attr_transaction():
    """
    Transactions must always have `amount`, `account`, `rmtinfo`,
    `creditor`.
    """
    basic_data = {
        'amount': 1,
        'account': acct_86,
        'creditor': biz,
    }
    for params in gen_dictionaries_with_one_missing_item(basic_data):
        payment = Payment(debtor=biz_with_cuc, account=acct_37)
        params['rmtinfo'] = 'Test'
        with pytest.raises(AttributeError):
            payment.add_transaction(**params)
            payment.xml()
    
    payment = simple_payment()
    with pytest.raises(AssertionError):
        payment.add_transaction(**basic_data)
        

def test_missing_CUC():
    "The `debtor` entity must always have a CUC."
    payment = Payment(debtor=biz, account=acct_37)
    payment.add_transaction(amount=1, account=acct_86, creditor=biz,
                            rmtinfo='Test')
    with pytest.raises(MissingCUCError):
        payment.xml()


def test_wrong_address_format():
    "Address must be a list or tuple of at most two lines."
    holder = IdHolder(
        name='Test Business S.P.A.', cf='12312312311',
        country='IT', address=['Xyz', 'Via Giuseppe Verdi, 15', '33100 Udine'],
        cuc='XYZZZZ')

    payment = Payment(debtor=holder, account=acct_37)
    payment.add_transaction(amount=1, account=acct_86, docs=(Invoice(1),),
                            creditor=alpha)
    with pytest.raises(AddressFormatError):
        payment.xml()
    holder.address = [
        'A single line which is too long to fit in the holder address '
        'field, which can hold at most 70 characters per line',
        'test'
    ]
    with pytest.raises(AddressFormatError):
        payment.xml()


def test_missing_abi():
    "If the debtor has a foreign account, the ABI must be supplied."
    payment = Payment(debtor=biz_with_cuc, account=foreign_acct)
    payment.add_transaction(amount=1, account=acct_86, rmtinfo='Test')
    with pytest.raises(MissingABIError):
        payment.xml()


def test_missing_bic():
    "If the creditor has a foreign account, the BIC must be supplied."
    payment = simple_payment()
    with pytest.raises(MissingBICError):
        payment.add_transaction(amount=1, account=foreign_acct, creditor=biz,
                                rmtinfo='Test')


def test_invalid_iban():
    payment = Payment(debtor=biz_with_cuc, account='ITINVALIDIBAN')
    with pytest.raises(InvalidIBANError):
        payment.xml()

    payment.account = 'XXINVALIDIBAN'
    payment.abi = '01234'
    with pytest.raises(InvalidIBANError):
        payment.xml()

    payment.account = acct_37.replace('IT37', 'IT99')
    with pytest.raises(InvalidIBANError):
        payment.xml()


def test_invalid_docs():
    payment = simple_payment()
    with pytest.raises(TypeError):
        payment.add_transaction(amount=100, account=acct_86, creditor=biz,
                                docs=[Invoice()])


def test_dup_e2eid():
    payment = simple_payment()
    payment.add_transaction(amount=1, creditor=alpha, account=acct_37,
                            eeid='Test1', rmtinfo='A')
    with pytest.raises(InvalidEndToEndIDError):
        payment.add_transaction(amount=2, creditor=beta, account=acct_37,
                                eeid='Test1', rmtinfo='B')



#!/usr/bin/python

"""
Import classes exposed as the API to consumer modules.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'

from .entity import IdHolder
from .payment import Payment
from .rmtinfo import Document, Invoice, CreditNote, DebitNote, Text
from .transaction import Transaction

from lxml import etree

PR_PREFIX = 'urn:CBI:xsd:CBIPaymentRequest.00.04.00'

etree.register_namespace('pr', PR_PREFIX)



import copy
from unidecode import unidecode
from datetime import date
from decimal import Decimal
from six import add_metaclass


class Ordered(object):
    def set_order(self, order):
        self.order = order


class BaseField(Ordered):
    """
    A field in a (fixed-size) text record.
    """
    @classmethod
    def newfield(cls, recordcls, pos, **fields):
        """
        Creates a field with the given name and length, then binds it
        with the given containing record or composite field.
        """
        if len(fields) > 1:
            raise Exception('Multiple k/w values not supported')
        items = list(fields.items())
        name = items[0][0]
        flen = items[0][1]
        instance = cls(pos, name, flen)
        recordcls.bind(instance)

    @classmethod
    def alt_newfield(cls, recordcls, flen, name, pos, **kwargs):
        """
        Alternate method to create and bind a field.
        """
        instance = cls(pos, name, flen, **kwargs)
        recordcls.bind(instance)

    def __init__(self, pos, name, flen, default=None):
        """
        Fields and records have fixed positions and sizes.
        Position, name, field length are the same for every record, so
        they belong to the descriptor.
        The stored value changes, so it belongs to the instance.
        """
        self.name = name
        self.pos = pos
        self._flen = flen

        if default is not None:
            self.__dict__['_default_value'] = default

    def __get__(self, obj, type=None):
        """
        Fields never store the original value; they just delegate the instance
        so that it can store their exported representation.
        """
        raise NotImplementedError('Error: self=%r, obj=%r, type=%r' % (self, obj, type))

    def __set__(self, obj, value):
        """
        Ask the instance to store the exported representation of the value.
        """
        obj.write_field(self.order, self.format(value))

    def format(self, value):
        """
        Format the value; raise an exception if the size is wrong, as every
        exported field has a fixed size.
        """
        s = self._specialized_format(value)
        if len(s) != self.size:
            raise Exception('Could not format properly value %r' % value)
        return s

    @property
    def size(self):
        "Return the exported field size in characters."
        return self._flen

    def get_default(self):
        "Return the exported representation of the default value."
        return [self.format(self._default_value)]

    slot_count = 1


class FieldContainerMeta(type):
    """
    A class containing a field, such as a record or a composite field,
    must perform this procedure upon initialization.
    """
    def __new__(*args, **kwargs):
        c = type.__new__(*args, **kwargs)
        c.pos = 1
        c.field_count = 0
        c.fields = []
        c.define_fields()
        c.set_defaults()
        return c


class FieldContainerMixin(object):
    "Common methods for all field containers (records and composite fields)."
    @classmethod
    def bind(cls, field):
        """
        Bind a field instance with this class.
        """
        if field.name in cls.__dict__:
            raise 'Already has a field with this name'
        #if field.pos != cls.pos:
        #   raise Exception('Offset error: field %r is defined at position %s,'
        #       ' but class %r is at position %s' % (field.name, field.pos,
        #           cls.__name__, cls.pos))
        setattr(cls, field.name, field)
        cls.fields.append(field)
        field.set_order(cls.field_count)
        cls.field_count += field.slot_count
        cls.pos += field.size

    @classmethod
    def set_defaults(cls):
        lofl = [f.get_default() for f in cls.fields]
        cls._defaults = [val for subl in lofl for val in subl]

    @classmethod
    def define_fields(cls):
        pass

    @classmethod
    def builders(cls, nf=None):
        class Builder(object):
            pass
        o = Builder()
        if nf is None:
            nf = lambda fcls: lambda *a, **kw: fcls.alt_newfield(cls, *a, **kw)
        o.an = nf(AlphaNumericField)
        o.nu = nf(NumericField)
        o.dc = nf(DecimalField)
        o.dt = nf(DateField)
        o.sn = nf(SNField)
        return o

    @classmethod
    def kw_builders(cls):
        return cls.builders(
            nf=lambda fcls: lambda *a, **kw:
            fcls.newfield(cls, *a, **kw))


class SubfieldAccessor(object):
    def __init__(self, itemaccessor, num):
        self.__dict__['record'] = itemaccessor.record
        self.__dict__['compositefield'] = itemaccessor.compositefield
        self.num = num

    def __setattr__(self, name, value):
        if name in self.compositefield.__class__.__dict__:
            field = self.compositefield.__class__.__dict__[name]
            offset = self.compositefield.field_count*self.num + \
                field.order + self.compositefield.order
            self.record.write_field(offset, field.format(value))
        else:
            object.__setattr__(self, name, value)


class ItemAccessor(object):
    def __init__(self, compositefield, record):
        self.compositefield = compositefield
        self.record = record

    def __getitem__(self, num):
        if num > self.compositefield.repetitions:
            raise IndexError
        return SubfieldAccessor(self, num)


@add_metaclass(FieldContainerMeta)
class CompositeField(FieldContainerMixin, Ordered):
    """
    A field that can be repeated and contain multiple compositefields.
    """

    __metaclass__ = FieldContainerMeta

    def __get__(self, obj, type=None):
        return ItemAccessor(self, obj)

    def __set__(self, obj, value):
        raise NotImplementedError

    def __init__(self, name, repetitions, offset):
        self.name = name
        self.repetitions = repetitions
        self.pos = offset
        self._total_size = sum([f.size for f in self.fields])*repetitions

    @property
    def slot_count(self):
        return self.field_count*self.repetitions

    @property
    def size(self):
        return self._total_size

    def get_default(self):
        lofl = self.repetitions*[f.get_default() for f in self.fields]
        m = [val for subl in lofl for val in subl]
        return m


@add_metaclass(FieldContainerMeta)
class BaseRecord(FieldContainerMixin):
    __metaclass__ = FieldContainerMeta

    def __init__(self):
        """
        Copy default values from class when creating instance.
        Then, prepare compositefield accessors.
        """
        self._values = copy.copy(self._defaults)

    def write_field(self, field, value):
        self._values[field] = value

    def format(self):
        return ''.join(self._values)

    def debug_format(self):
        return '%r' % self._values


##########################
# Useful implementations #
##########################


class AlphaNumericField(BaseField):
    def _specialized_format(self, value):
        r = unidecode(value).strip().replace(
            '\x0d\x0a', '\x0a').replace('\x0a', ' / ').ljust(self._flen)
        if len(r) > self._flen:
            r = r[:self._flen]
        return r

    _default_value = u''


class NumericField(BaseField):
    def _specialized_format(self, value):
        if value is None:
            return u' '*self._flen
        return str(int(value)).zfill(self._flen)

    _default_value = None


class SNField(BaseField):
    def _specialized_format(self, value):
        v = bool(value)
        if v:
            return 'S'
        else:
            return 'N'

    _default_value = False


class DateField(BaseField):
    def __init__(self, *args, **kwargs):
        super(DateField, self).__init__(*args, **kwargs)
        if self._flen != 8:
            raise Exception('Invalid length for DateField')

    def _specialized_format(self, value):
        if value is None:
            return '00000000'
        elif isinstance(value, date):
            return value.strftime('%d%m%Y')
        else:
            raise Exception('Invalid type for date: %s' % type(value))

    _default_value = None


class CBIDateField(BaseField):
    def __init__(self, *args, **kwargs):
        super(CBIDateField, self).__init__(*args, **kwargs)
        if self._flen != 6:
            raise Exception('Invalid length for CBIDateField: %s' % self._flen)

    def _specialized_format(self, value):
        if value is None:
            return '      '
        elif hasattr(value, 'strftime'):
            return value.strftime('%d%m%y')
        else:
            raise Exception('Invalid type for date: %r' % value)

    _default_value = None

class DecimalField(BaseField):
    def __init__(self, len, name, cint, cdec, pos):
        if cint+cdec != len:
            raise Exception('Invalid lengths for field')
        super(DecimalField, self).__init__(pos, name, len)
        self.cint = cint
        self.cdec = cdec
        self.multiplier = Decimal(10**self.cdec)

    @classmethod
    def alt_newfield(cls, recordcls, *args):
        instance = cls(*args)
        recordcls.bind(instance)

    def _specialized_format(self, value):
        if value is None:
            value = Decimal(0)
        return str((value * self.multiplier).to_integral()).zfill(self._flen)

    _default_value = None


class CBIDecimalField(DecimalField):
    def __init__(self, pos, name, flen, default=None):
        super(CBIDecimalField, self).__init__(flen, name, flen-2, 2, pos)

######################
# Examples and tests #
######################


class ExampleCompositeField(CompositeField):
    @classmethod
    def define_fields(cls):
        f = lambda *a, **k: AlphaNumericField.newfield(cls, *a, **k)
        f(31, hi=3)
        f(35, earth=5)


class ExampleRecord(BaseRecord):
    @classmethod
    def define_fields(cls):
        f = cls.kw_builders()
        f.an(1, hello=10)
        f.an(11, world=20)
        cls.bind(ExampleCompositeField('repeat', 5, 31))
        f.an(71, ending=8)


if __name__ == '__main__':
    r = ExampleRecord()
    r.world = u'myWorld'
    r.repeat[0].hi = u'a'
    r.repeat[0].earth = u'b'
    r.repeat[1].hi = u'c'
    r.repeat[4].hi = u'm'
    r.ending = u'mw'
    print(r.debug_format())
    print('%r' % r.format())




from .records import BaseRecord, AlphaNumericField, NumericField, \
    CBIDateField, CBIDecimalField


class CBIRecord(BaseRecord):
    @classmethod
    def builders(cls, nf=None):
        def nnf(fcls):
            def start(*a, **kw):
                m = fcls(*a, **kw)
                cls.bind(m)
            return start
        if nf is None:
            nf = nnf
        o = super(CBIRecord, cls).builders(nf=nf)
        o.dt = nf(CBIDateField)
        #     def __init__(self, len, name, cint, cdec, pos):
        o.cur = nf(CBIDecimalField)
        return o


class PCRecord(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'PC')
        f.an(4, 'sender', 5)
        f.nu(9, 'recipient', 5)
        f.dt(14, 'creation', 6)
        f.an(20, 'name', 20)
        f.an(40, 'available', 6)
        f.an(46, 'filler2', 59)
        f.an(105, 'flow_qualifier', 7)
        f.an(112, 'filler3', 1)
        f.an(113, 'prio', 1)
        f.an(114, 'currency_code', 1, default=u'E')
        f.an(115, 'filler4', 1)
        f.an(116, 'not_available', 5)


class EFRecord(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'EF')
        f.an(4, 'sender', 5)
        f.nu(9, 'recipient', 5)
        f.dt(14, 'creation', 6)
        f.an(20, 'name', 20)
        f.an(40, 'available', 6)
        f.nu(46, 'orders', 7)
        f.cur(52, 'negative_amounts', 15)
        f.cur(68, 'positive_amounts', 15)
        f.nu(83, 'records', 7)
        f.an(90, 'filler2', 23)
        f.an(113, 'prio', 1)
        f.an(114, 'currency_code', 1, default=u'E')
        f.an(115, 'not_available', 6)

# Record 10

class TransferInfo(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'10')
        f.nu(4, 'prog_number', 7)
        f.an(11, 'filler2', 6)
        f.dt(17, 'execution_date', 6)
        f.dt(23, 'payee_date', 6)
        f.an(29, 'purpose', 5)
        f.cur(34, 'amount', 13)
        f.an(47, 'sign', 1, default=u'+')
        f.nu(48, 'ord_abi', 5)
        f.nu(53, 'ord_cab', 5)
        f.an(58, 'ord_account', 12)
        f.nu(70, 'rec_abi', 5)
        f.nu(75, 'rec_cab', 5)
        f.nu(80, 'rec_account', 12)
        f.an(92, 'filler3', 22)
        f.nu(114, 'payment_method', 1, default=1)
        f.an(115, 'filler4', 4)
        f.an(119, 'prio', 1)
        f.an(120, 'currency_code', 1, default=u'E')


class PayerIBANInfo(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'16')
        f.nu(4, 'prog_number', 7)
        f.an(11, 'iban', 27)
        f.an(38, 'filler2', 83)


class PayeeIBANInfo(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'17')
        f.nu(4, 'prog_number', 7)
        f.an(11, 'iban', 27)
        f.an(38, 'filler2', 83)


class PayerInfo(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'20')
        f.nu(4, 'prog_number', 7)
        f.an(11, 'name', 30)
        f.an(41, 'address', 30)
        f.an(71, 'city', 30)
        f.an(101, 'tax_code', 16)
        f.an(117, 'filler2', 4)


class PayeeInfo(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'30')
        f.nu(4, 'prog_number', 7)
        f.an(11, 'name', 90)        
        f.an(101, 'tax_code', 16)
        f.an(117, 'filler2', 4)


class PayeeAddress(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'40')
        f.nu(4, 'prog_number', 7)
        f.an(11, 'address', 30)
        f.an(41, 'postal_code', 5)
        f.an(46, 'town', 25)
        f.an(71, 'bank', 50)


class PurposeInfo(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'50')  # may become 60
        f.nu(4, 'prog_number', 7)
        f.an(11, 'desc', 90)
        f.an(101, 'filler2', 20)


class StatusRequest(CBIRecord):
    @classmethod
    def define_fields(cls):
        f = cls.builders()
        f.an(1, 'filler', 1)
        f.an(2, 'record_type', 2, default=u'70')
        f.nu(4, 'prog_number', 7)
        f.an(11, 'filler2', 15)
        f.an(26, 'not_available', 5)
        f.an(31, 'flow_qualifier', 7)
        f.an(38, 'mp_code', 5)
        f.an(43, 'filler3', 27)
        f.an(70, 'status_request_flag', 1)
        f.an(71, 'unique_code', 30)
        f.an(101, 'filler4', 10)
        f.an(111, 'payee_cin', 1)
        f.an(112, 'filler5', 1)
        f.an(113, 'check_keys', 8)





#!/usr/bin/python

"""
This module deals with the information related to a single transaction.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'

from lxml import etree
from decimal import Decimal
from .util import AttributeCarrier
from .bank import Bank
from .account import Account
from .cbibon_dom import TransferInfo, PayerIBANInfo, PayeeIBANInfo, \
    PayerInfo, PayeeInfo, PayeeAddress, PurposeInfo, StatusRequest
import sys

if sys.version_info[0] >= 3:
    # pylint: disable=redefined-builtin
    # pylint: disable=invalid-name
    basestring = str


# Categories to CBI purpose mapping:

CATEGORY_CBI_MAP = {
    'SALA': '27020',
    'PENS': '27010',
    'SUPP': '48000',
}


class MissingBICError(Exception):
    """
    Raised when a BIC code is needed but not specified.
    """

class Transaction(AttributeCarrier):
    """
    Transaction describes a single credit transfer from the debtor's account
    (specified in the global payment data) to a creditor account.
    """
    # pylint: disable=no-member

    allowed_args = (
        'tx_id', 'eeid', 'category', 'rmtinfo', 'amount',
        'ultimate_debtor', 'bic', 'account', 'creditor', 'ultimate_creditor',
        'docs', 'purpose', 'payment_seq', 'payment_id',
        'register_eeid_function', 'payment', 'cbi_purpose')

    def __init__(self, *args, **kwargs):
        self.purpose = 'SUPP'
        self.eeid_registered = False
        super(Transaction, self).__init__(*args, **kwargs)

    def gen_id(self):
        "Generate a sequential ID, if not supplied, for the `InstrId` element."
        self.tx_id = str(self.payment_seq)

    def gen_eeid(self):
        "Generate a unique ID for the `EndToEndId` element."
        self.eeid = '%s-%06d' % (self.payment_id, self.payment_seq)

    def perform_checks(self):
        "Check lengths and types for the attributes."
        # pylint: disable=access-member-before-definition
        # pylint: disable=attribute-defined-outside-init
        if not hasattr(self, 'tx_id'):
            self.gen_id()
        self.max_length('tx_id', 35)

        if not hasattr(self, 'eeid'):
            self.gen_eeid()
        self.max_length('eeid', 35)
        if not self.eeid_registered:
            self.register_eeid_function(self.eeid)
            self.eeid_registered = True

        self.length('purpose', 4)
        if not hasattr(self, 'category'):
            self.category = self.purpose

        self.length('category', 4)

        if not isinstance(self.amount, Decimal):
            self.amount = Decimal(str(self.amount)).quantize(Decimal('.01'))
        if isinstance(self.account, basestring):
            self.account = Account(iban=self.account)
        if self.account.is_foreign():
            if not hasattr(self, 'bic'):
                raise MissingBICError
            bic_length = len(self.bic)
            assert bic_length in (8, 11)

        if hasattr(self, 'rmtinfo'):
            assert not hasattr(self, 'docs')
            assert len(self.rmtinfo) <= 140

        if hasattr(self, 'docs'):
            for item in self.docs:
                assert hasattr(item, '__str__')

        assert hasattr(self, 'docs') or hasattr(self, 'rmtinfo')

    def emit_tag(self):
        """
        Returns the XML tag for the transaction.
        """
        root = etree.Element('CdtTrfTxInf')
        pmtid = etree.SubElement(root, 'PmtId')
        etree.SubElement(pmtid, 'InstrId').text = self.tx_id
        etree.SubElement(pmtid, 'EndToEndId').text = self.eeid
        info = etree.SubElement(root, 'PmtTpInf')
        purp = etree.SubElement(info, 'CtgyPurp')
        etree.SubElement(purp, 'Cd').text = self.category
        amt = etree.SubElement(root, 'Amt')
        etree.SubElement(
            amt, 'InstdAmt', attrib={'Ccy': 'EUR'}).text = str(self.amount)
        if hasattr(self, 'ultimate_debtor'):
            root.append(self.ultimate_debtor.__tag__('UltmtDbtr'))
        if self.account.is_foreign():
            agt = etree.SubElement(root, 'CdtrAgt')
            agt.append(Bank(bic=self.bic).__tag__())
        root.append(self.creditor.__tag__('Cdtr'))
        root.append(self.account.__tag__('CdtrAcct'))
        if hasattr(self, 'ultimate_creditor'):
            root.append(self.ultimate_creditor.__tag__('UltmtCdtr'))
        rmtinf = etree.SubElement(root, 'RmtInf')
        if hasattr(self, 'rmtinfo'):
            etree.SubElement(rmtinf, 'Ustrd').text = self.rmtinfo
        else:
            etree.SubElement(rmtinf, 'Ustrd').text = ''.join(
                [str(doc) for doc in self.docs])
        return root

    def cbi_records(self, prog=None):
        if self.account.is_foreign():
            raise Exception('Cannot use a foreign IBAN with CBI text files')

        records = []

        xinfo = TransferInfo()
        xinfo.prog_number = prog
        if hasattr(self.payment, 'execution_date'):
            xinfo.execution_date = self.payment.execution_date
        # TODO: allow generic codes for salaries, pensions
        if hasattr(self, 'cbi_purpose'):
            xinfo.purpose = self.cbi_purpose
        elif hasattr(self, 'category'):
            if self.category in CATEGORY_CBI_MAP:
                xinfo.purpose = CATEGORY_CBI_MAP[self.category]
            else:
                raise Exception('Cannot map cateogry %r; please supply the '
                                '\'cbi_purpose\' attribute' % self.category)
        else:
            xinfo.purpose = '48000'

        xinfo.amount = self.amount
        xinfo.ord_abi = self.payment.bank.abi
        xinfo.ord_cab = self.payment.cab
        xinfo.ord_account = self.payment.cc
        if hasattr(self.payment, 'high_priority'):
            if self.payment.high_priority:
                xinfo.prio = 'U'
        records.append(xinfo.format())

        ord_iban = PayerIBANInfo()
        ord_iban.prog_number = prog
        ord_iban.iban = self.payment.account.iban
        records.append(ord_iban.format())

        ben_iban = PayeeIBANInfo()
        ben_iban.prog_number = prog
        ben_iban.iban = self.account.iban
        records.append(ben_iban.format())

        ord_info = PayerInfo()
        ord_info.prog_number = prog
        ord_info.name = self.payment.debtor.name
        ord_info.tax_code = self.payment.debtor.cf
        records.append(ord_info.format())

        ben_info = PayeeInfo()
        ben_info.prog_number = prog
        ben_info.name = self.creditor.name
        if hasattr(self.creditor, 'cf'):
            ben_info.tax_code = self.creditor.cf
        records.append(ben_info.format())

        # Record 40 is not written
        records += self.rmt_cbi_records(prog=prog)

        # The status request record is empty
        status_req = StatusRequest()
        status_req.prog_number = prog
        records.append(status_req.format())

        return records

    @classmethod
    def rmtinfo_record(cls, record_type, prog, line):
        record = PurposeInfo()
        record.prog_number = prog
        record.record_type = record_type
        record.desc = line
        return record

    def rmt_cbi_records(self, prog):
        records = []
        if hasattr(self, 'rmtinfo'):
            if len(self.rmtinfo) <= 90:
                record = PurposeInfo()
                record.prog_number = prog
                record.desc = self.rmtinfo
                records += [record.format()]
            else:
                start = 0
                while start < len(self.rmtinfo):
                    records += [self.rmtinfo_record('60', prog,
                        self.rmtinfo[start:start+90]).format()]
        else:
            record_type = '60'
            if len(self.docs) <= 3:
                record_type = '50'
            i = 0
            line = ''
            while i < len(self.docs):
                line += self.docs[i].cbi()
                if i % 3 == 2:
                    records.append(self.rmtinfo_record(record_type, prog, line
                        ).format())
                    line = ''
                i += 1
            if line != '':
                records.append(self.rmtinfo_record(record_type, prog, line
                    ).format())
            if len(records) > 5:
                raise Exception('Too many documents for remittance info')
        return records


            








#!/usr/bin/python

"""
This module automatically prepares unstructured remittance information
according to the EACT standard.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'


from decimal import Decimal
from .util import AttributeCarrier


class Document(AttributeCarrier):
    """
    The Document class represents a financial document that can be mentioned
    in the unstructured remittance information of a transaction.

    Please see the EACT Standard for Remittance Information:
    http://www.eact.eu/main.php?page=SEPA
    """

    tag = '/DOC/'

    def __init__(self, number, amount=None, date=None):
        """
        Argument processing is not based on keyword parameters, so we can't call
        `AttributeCarrier`'s `__init__()` method.
        """
        # pylint: disable=super-init-not-called
        self.number = number
        if amount:
            self.amount = amount
        if date:
            self.date = date

    def perform_checks(self):
        "Check the document number and, optionally, the date and amount."
        assert int(self.number) is not None
        if hasattr(self, 'date'):
            assert hasattr(self.date, 'strftime')
            assert hasattr(self, 'amount')
        if hasattr(self, 'amount'):
            if not isinstance(self.amount, Decimal):
                self.amount = Decimal(str(self.amount))

    def __str__(self):
        self.perform_checks()
        items = [str(self.number)]
        if hasattr(self, 'amount'):
            items += [str(self.amount.quantize(Decimal('.01')))]
        if hasattr(self, 'date'):
            items += [self.date.strftime('%Y%m%d')]
        return self.tag + '/ '.join(items)

    def cbi(self):
        # CBI-BON-001, record 50/60
        if hasattr(self, 'number') and hasattr(self, 'date'):
            return '%24s%s' % (self.number, self.date.strftime('%d%m%Y'))
        elif hasattr(self, 'number'):
            return '%24s      ' % (self.number)
        else:
            raise Exception('Invalid data for CBI format')


class Invoice(Document):
    "Commercial invoice."
    tag = '/CINV/'


class CreditNote(Document):
    "Credit note."
    tag = '/CREN/'


class DebitNote(Document):
    "Debit note."
    tag = '/DEBN/'


class Text(Document):
    "Free text."
    # pylint: disable=super-init-not-called
    tag = '/TXT/'

    def __init__(self, text):
        self.text = text

    def perform_checks(self):
        assert hasattr(self, 'text')

    def __str__(self):
        # pylint: disable=no-member
        self.perform_checks()
        return self.tag + self.text

    def cbi(self):
        return '%30s' % self.text



#!/usr/bin/python
#pylint: disable=no-member

"""
Bank class.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'

from .util import AttributeCarrier
from lxml import etree
import re

ABI_RE = re.compile(r'\d{5}')


class Bank(AttributeCarrier):
    """
    Represents a bank. It can show either a BIC code (for creditor banks)
    or an ABI code (for the debtor bank).
    """
    allowed_args = ('bic', 'abi')

    def perform_checks(self):
        'Checks that the BIC or the ABI are present, and that they are valid.'
        if hasattr(self, 'bic'):
            code_length = len(self.bic)
            # TODO: more formal checks (ISO 9362)
            assert code_length in (8, 11)
        if hasattr(self, 'abi'):
            assert ABI_RE.match(self.abi) is not None

    def emit_tag(self, output_abi=False):
        "Returns a XML tree for the bank, optionally providing the ABI code."
        root = etree.Element('FinInstnId')
        if hasattr(self, 'bic'):
            etree.SubElement(root, 'BIC').text = self.bic
        if output_abi:
            clearing = etree.SubElement(root, 'ClrSysMmbId')
            etree.SubElement(clearing, 'MmbId').text = self.abi
        return root



from sepacbi import Payment, Invoice, DebitNote, Text
from .definitions import *
import re
import sys
from datetime import datetime, date

from lxml import etree

PYTHON3 = False
if sys.version_info[0] >= 3:
    PYTHON3 = True

def canonicalize_cbi(text):
    today = date.today().strftime('%d%m%Y')
    return text.replace(today, '999999')

def compare_cbi(text, filename, save=False):
    """
    Compare a generated XML tree with the content of a specific file.
    """
    from os.path import dirname, join
    full_path = join(dirname(__file__), filename)
    if save:
        open(full_path, 'w').write(canonicalize_cbi(text))
        #raise Exception('saved!')
    file_content = open(full_path, 'r').read()
    assert text == file_content


def test_payment_basic():
    payment = Payment(debtor=biz_with_sia, account=acct_37, req_id='StaticId',
                      execution_date=date(2014, 5, 15))
    payment.add_transaction(amount=198.25, account=acct_86, creditor=beta,
                            rmtinfo='Causale 1')
    compare_cbi(payment.cbi_text(), 'payment_basic.txt', save=True)


def test_payment_multitrans():
    payment = Payment(debtor=biz_with_sia, account=acct_37, req_id='StaticId',
                      execution_date=date(2014, 5, 15))
    payment.add_transaction(amount=198.25, account=acct_86, creditor=beta,
                            rmtinfo='Causale 1')
    payment.add_transaction(amount=350.00, account=acct_37,
                            creditor=biz_with_sia,
                            rmtinfo='Altra causale')
    payment.add_transaction(amount=9532.21, account=acct_86,
                            bic='ABCDESNN', creditor=alpha,
                            docs=[Invoice(18512, 4500),
                                  DebitNote(1048, 5032.21,
                                            date(1995, 4, 21))])
    payment.add_transaction(
        amount=1242.80, creditor=pvt, account=acct_86,
        category='SALA', rmtinfo='Salary')

    compare_cbi(payment.cbi_text(), 'payment_multitrans.txt')


def test_payment_misc_features():
    payment = Payment(
        debtor=biz, account=acct_37, req_id='StaticId',
        execution_date=date(2014, 5, 15),
        ultimate_debtor=beta, charges_account=acct_86, envelope=True,
        initiator=biz_with_sia, batch=True, high_priority=True)
    payment.add_transaction(amount=198.25, account=acct_86, creditor=beta,
                            rmtinfo='Causale 1')
    compare_cbi(payment.cbi_text(), 'payment_misc_1.txt')

    payment.high_priority = False
    payment.add_transaction(amount=350.00, account=acct_37,
                            creditor=biz_with_cuc,
                            rmtinfo='Altra causale')
    payment.add_transaction(amount=9532.21, account=acct_86,
                            bic='ABCDESNN', creditor=alpha,
                            docs=[Invoice(18512, 4500, date(2009, 7, 17)),
                                  DebitNote(1048, 5032.21,
                                            date(1995, 4, 21))])
    payment.add_transaction(
        amount=1242.80, creditor=pvt, account=acct_86,
        category='SALA', docs=[Text('Salary payment')])

    compare_cbi(payment.cbi_text(), 'payment_misc_2.txt')



from copy import copy

from sepacbi import IdHolder


biz = IdHolder(name='Test Business S.P.A.', cf='12312312311',
               country='IT', address=['Via Giuseppe Verdi, 15', '33100 Udine'])

pvt = IdHolder(name='Mario Rossi', cf='RSSMRA50C10F842H',
               country='IT', private=True)

biz_with_cuc = copy(biz)
biz_with_cuc.cuc = 'S0215325Z'
biz_with_sia = copy(biz)
biz_with_sia.sia_code = '0A123'
acct_86 = 'IT 86 U 07601 11500 000010117463'
acct_37 = 'IT37Z0760101600000028426203'
foreign_acct = 'ES38 2100 1579 8002 0025 5488'

alpha = IdHolder(
    name='Alpha s.r.l.', cf='IT01234567890', country='IT',
    address=('Piazza Rossini, 8/A', '44044 Xyz'))
beta = IdHolder(name='Beta s.n.c.', code='ESQ01231244')



#!/usr/bin/python

"""
Setup routines.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'

from setuptools import setup, find_packages

setup(
    name="sepacbi",
    version="0.1.7",
    description="SEPA Credit Transfer request (CBI) XML generator",
    long_description="{0:s}\n\n".format(
        open("README.rst").read()
    ),
    author="Emanuele Pucciarelli",
    author_email="ep@corp.it",
    url="http://github.com/puccia/sepacbi/",
    download_url="http://github.com/puccia/sepacbi/",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Office/Business :: Financial",
    ],
    license="3-clause BSD",
    keywords="xml finance banking payments",
    platforms="All",
    packages=find_packages("."),
    install_requires=['lxml', 'six', 'unidecode'],
    zip_safe=True
)



#!/usr/bin/python

"""
Bank accounts.
"""

__copyright__ = 'Copyright (c) 2014 Emanuele Pucciarelli, C.O.R.P. s.n.c.'
__license__ = '3-clause BSD'

from .util import AttributeCarrier
from lxml import etree


class Account(AttributeCarrier):
    """
    The Account class represents a bank account, identified by an IBAN.
    """
    allowed_args = ('iban',)

    def __init__(self, *args, **kwargs):
        self.iban = ''
        super(Account, self).__init__(*args, **kwargs)
        self.clean_iban()

    def clean_iban(self):
        "Canonicalize the IBAN."
        self.iban = self.iban.upper().replace(' ', '')

    def is_foreign(self):
        """
        Return a boolean value telling whether the IBAN designates a foreign
        account (from the viewpoint by the Italian banking system).
        """
        return self.iban[:2] not in ('IT', 'SM')

    def perform_checks(self):
        "Validates the IBAN."
        from . import iban
        self.clean_iban()
        iban.validate(self.iban)

    def emit_tag(self, tag):
        "Emit a XML tag using the provided string for the root element."
        root = etree.Element(tag)
        acct_id = etree.SubElement(root, 'Id')
        etree.SubElement(acct_id, 'IBAN').text = self.iban
        return root
